/**
 * MapUtils.js
 *
 * MapUtil contains all the logic to build a set of entities to be displayed on a map for any event type.
 * This code is platform independent and should not contain any logic or conditions which rely on knowledge
 * of the platform it is running on. Furthermore the code should be generic and not have special casses
 * for different event types (run, hike, bike, etc) either.
 *
 * The main entry point for MapUtils is the generateEntities method. If at any point the signature for this
 * method changes, updates will need to be made on each platform, so please coordinate with the appropriate
 * engineers on each platform if a change is needed.
 *
 *       ---------- IMPORTANT ----------
 *
 *       All platforms (Web, iOS, Windows and Android) will be utilizing and sharing this script so efforts
 *       must be made to keep it generic and sharable.
 *
 *       Never modify this script outside of the Shared directory. All changes should be made in the KDK
 *       branch at Source/Shared/BingMap/MapUtils.js and then applied to the different platforms.
 *
 *       To deploy changes to mobile platforms, run Source/Shared/CopyAssetsToPlatform.ps1 with powershell
 *       To deploy changes to the web merge the changes to the health-web repo under /Scripts/third-party/MapUtils.js
 *
 *       ALWAYS deploy changes to all platforms and loop in engineers from each to review modifications.
 *       ------------------------------
 */
function MapUtils(assetPath) {
    this.constants = {
        'latLongUnitsPerDegree': 10000000,
        'approx100FeetInDegrees': 0.0004,
        'checkAheadForNearbyPoints': 5,
        'scaledPaceInterval': 33,
        'paceColors': ['#0083B2', '#FFB900', '#EB1123'],
        'pausedColor': '#000000',
        'nilColor': '#777777',
        'strokeThickness': 5,
        'wayPoint': 'WayPoint',
        'splitMarker': 'Split',
        'startMarker': 'Start',
        'pauseMarker': 'Pause',
        'endMarker': 'End',
        'poiMarker': 'PointOfInterest',
        'speedThreshold': 30,
        'gpsErrorThreshold': 3000,
        'pushPinHeight': 46,
        'pushPinWidth': 25,
        'pushPinHeightFlag': 31,
        'pushPinWidthFlag': 21,
        //Match to height & width if they change!
        'pushPinAnchorPoint': new Microsoft.Maps.Point(25 * 0.5, 46 * 0.9),
        'pushPinAnchorPointFlag': new Microsoft.Maps.Point(21 * 0.5, 31 * 1.0),
        'enableDebug': false,
        'userGeneratedID': "UserGenerated",

        poiImages: {
            UserGenerated: { img: "POI_flags-O1.png" },
            ElevationMax: { img: "POI_flags_max_elevation.png" },
            ElevationMin: { img: "POI_flags_min_elevation.png" },
            TimeMidPoint: { img: "POI_flags_midpoint.png" },
            Pause: { img: "POI_flags_pause.png" },
            PauseAuto: { img: "POI_flags_pause.png" },
            Sunrise: { img: "POI_flags_sunset.png" },
            Sunset: { img: "POI_flags_sunset.png" },
            Start: { img: "POI_flags_start.png" },
            End: { img: "POI_flags_finish.png" }
        }
    };

    this.paceColorSize = this.constants.paceColors.length;
    this.assetPath = assetPath;

    this.getPaceSegment = function (pace) {
        var segment = parseInt(pace / this.constants.scaledPaceInterval, 10);

        if (segment > this.paceColorSize) {
            segment = this.paceColorSize - 1;
        }

        return segment;
    };

    this.getColorForPace = function (pace) {
        return this.constants.paceColors[this.getPaceSegment(pace)];
    };

    this.getColorForPoint = function (point) {
        if (point == null) {
            return this.constants.nilColor;
        } else if (point.IsPaused) {
            return this.constants.pausedColor;
        } else {
            return this.constants.paceColors[this.getPaceSegment(point.ScaledPace)];
        }
    };

    this.arePacesInTheSameZone = function (previousPace, currentPace) {
        if (point == null) {
            return false;
        }
        return this.getPaceSegment(previousPace) === this.getPaceSegment(currentPace);
    };

    this.arePointsInTheSameZone = function (previousPoint, currentPoint) {
        if (previousPoint.IsPaused !== currentPoint.IsPaused) {
            return false;

        } else {
            return this.getPaceSegment(previousPoint.ScaledPace) === this.getPaceSegment(currentPoint.ScaledPace);
        }
    };

    this.createLocation = function (point) {
        if (point.hasOwnProperty('Location')) {
            var location = point.Location;

            // TODO: Temporary workaround until we can make serializer serialize nil object.
            if (location != null && location.hasOwnProperty('Latitude') && location.hasOwnProperty('Longitude')) {
                var latitude = location.Latitude / this.constants.latLongUnitsPerDegree;
                var longitude = location.Longitude / this.constants.latLongUnitsPerDegree;

                return new Microsoft.Maps.Location(latitude, longitude);
            }
        }

        return null;
    }

    this.createPushPin = function (location, icon, isFlag, text, typeName) {
        return new Microsoft.Maps.Pushpin(location,
            {
                text: text,
                icon: this.assetPath + icon,
                height: isFlag ? this.constants.pushPinHeightFlag : this.constants.pushPinHeight,
                width: isFlag ? this.constants.pushPinWidthFlag : this.constants.pushPinWidth,
                anchor: isFlag ? this.constants.pushPinAnchorPointFlag : this.constants.pushPinAnchorPoint,
                typeName: typeName
            });
    }

    /**
     * Generates the array of MapPoint PushPins (entities) used to draw the path on the map.
     * This function is shared between apps and web and should be UI agnostic.
     * @param routePointsFullRoute: Pass in an empty array that will be added to by this function. It's then used by the caller to show the correct zoom levelon the map
     * @param mapPoints: an array of MapPoints that are used to draw the map
     * @param shouldHideMarkers: boolean that indicates if split markers are hidden or not. Defaults to false
     * @param customSplit: int can be used to indicate a custom split. Defaults to 1
     * @param ignoreGpsLoss: boolean that indicates if all points should be connected even if GPS is lost. Defaults to true
     * @param enableFiltering: boolean that indicates if points should be filtered out if they are not strong enough. Defaults to true
     * @param pathHexColor: string hex color (e.g. "#ffffff") that should be used for the path on the map. If not specified, it will color the path based on pace
     * @param useMarkerForEnds: boolean that indicates if the start and end shoudl use pins with flag icons. If not, it will use green and red dots. Defaults to false
     */
    this.generateEntities = function (routePointsFullRoute, mapPoints, shouldHideMarkers, customSplit, ignoreGpsLoss, enableFiltering, pathHexColor, useMarkerForEnds) {
        // set default parameter values
        shouldHideMarkers = shouldHideMarkers == null ? false : shouldHideMarkers;
        customSplit = customSplit == null ? 1 : customSplit;
        ignoreGpsLoss = ignoreGpsLoss == null ? true : ignoreGpsLoss;
        enableFiltering = enableFiltering == null ? true : enableFiltering;
        useMarkerForEnds = useMarkerForEnds == null ? false : useMarkerForEnds;

        if (mapPoints == null) {
            return false;
        }

        // 2nd array for continuous sections, so maps doesn't auto-connect discontinuous sections where GPS was lost
        var routePointsSection = new Array();
        var entities = new Array();

        var startIndex = -1;
        var previousPoint = null;
        var addedStartMarker = false;
        var lastValidLocation = null;

        var filteredMapPoints = this.filterMapPoints(mapPoints, ignoreGpsLoss, enableFiltering);

        var splitMarkerIndex = 0;
        var userPoiMarkerIndex = 0;
        var pin;
        for (var i = 0; i < filteredMapPoints.length; i++) {
            var point = filteredMapPoints[i];
            var isMarker = false;

            var location = this.createLocation(point);

            if (location) {
                if (!addedStartMarker) {
                    addedStartMarker = true;
                    // This is a start marker.  Always goes on the First Valid GPS point.
                    // A final Split time pin may overlap and cover the End pin.
                    // Most real world runs will be a Loop, where Start and End are the same location (ie Home.)
                    // Save and add the startPin and endPin last after the for loop finishes.
                    if (useMarkerForEnds) {
                        pin = this.createPushPin(location, this.constants.poiImages.Start.img, true, null, "icon start");
                    } else {
                        pin = this.createPushPin(location, 'start.png', false);
                    }
                    entities.push(pin);

                    isMarker = true;
                } else if (point.MapPointType === this.constants.splitMarker) {
                    splitMarkerIndex = point.SplitOrdinal == null ? splitMarkerIndex + 1 : point.SplitOrdinal + 1;

                    if (splitMarkerIndex % customSplit === 0) {
                        if (!shouldHideMarkers) {
                            pin = this.createPushPin(location, 'pin.png', false, splitMarkerIndex.toString());
                            entities.push(pin);

                            isMarker = true;
                        } else {
                            pin = this.createPushPin(location, 'dot.png', false);
                            entities.push(pin);

                            isMarker = true;
                        }
                    }
                } else if (point.MapPointType === this.constants.poiMarker && point.POIType === this.constants.userGeneratedID) {
                    userPoiMarkerIndex = point.UserPOIOrdinal == null ? userPoiMarkerIndex + 1 : point.UserPOIOrdinal;
                    pin = this.createPushPin(location, this.constants.poiImages.UserGenerated.img, true, userPoiMarkerIndex.toString());
                    entities.push(pin);

                    isMarker = true;
                } else if (point.MapPointType === this.constants.poiMarker) {
                    var poiImg = null;
                    if (this.constants.poiImages[point.POIType]) {
                        poiImg = this.constants.poiImages[point.POIType].img;
                    }

                    if (poiImg) {
                        pin = this.createPushPin(location, poiImg, true, null, "icon " + point.POIType);
                        entities.push(pin);

                        isMarker = true;
                    }
                } else if (point.IsPaused &&
                    (!previousPoint || !previousPoint.IsPaused || !previousPoint.Location) &&
                    point.MapPointType !== this.constants.endMarker && point.MapPointType !== this.constants.startMarker) {
                    // To render a pause marker, it has to meeting the following conditions:
                    // 1. The point.IsPaused = true.
                    // 2. The previous point doesn't exist, or if it exists, it's not paused, or if it exists and is paused but has no location.
                    // 3. The point type is not start or end marker.
                    if (pathHexColor) {
                        pin = this.createPushPin(location, this.constants.poiImages.Pause.img, true, null, "icon pause");
                    } else {
                        pin = this.createPushPin(location, 'pause.png', false);
                    }
                    entities.push(pin);

                    isMarker = true;
                }
            }

            var shouldCompleteSegment = false;
            var isRendering;
            if (location) {
                isRendering = startIndex !== -1;
                if (isRendering) {
                    // We are plotting the route right now and this is a valid GPS point.
                    routePointsFullRoute.push(location);
                    routePointsSection.push(location);

                    shouldCompleteSegment = !this.arePointsInTheSameZone(previousPoint, point);
                } else {
                    // We are not plotting the route right now and this is a valid GPS point.
                    routePointsFullRoute.push(location);
                    routePointsSection.push(location);

                    startIndex = routePointsSection.length - 1;
                }
            } else {
                // This is not a valid GPS point
                shouldCompleteSegment = isRendering;
            }

            if (shouldCompleteSegment || i === filteredMapPoints.length - 1) {
                // Get the color from the point.
                var strokeColor = pathHexColor || this.getColorForPoint(previousPoint);

                // We want to take startIndex - 1 because we want the segments to overlap so that
                // it doesn't look disconnected when zoomed in.
                var route = new Microsoft.Maps.Polyline(
                    routePointsSection.slice(startIndex <= 0 ? 0 : startIndex - 1),
                    {
                        strokeColor: Microsoft.Maps.Color.fromHex(strokeColor),
                        strokeThickness: this.constants.strokeThickness
                    });

                entities.push(route);

                //Start a new route section; with this point in it to overlap to not look disconnected
                routePointsSection = new Array();

                if (location) {
                    routePointsSection.push(location);
                }

                startIndex = 0;
            }

            previousPoint = point;

            if (location && !isMarker) {
                lastValidLocation = location;
            }
        }

        //Add End marker on the last GPS point
        if (lastValidLocation) {
            if (useMarkerForEnds) {
                pin = this.createPushPin(lastValidLocation, this.constants.poiImages.End.img, true, null, "icon end");
            } else {
                pin = this.createPushPin(lastValidLocation, 'end.png', false);
            }
            entities.push(pin);
        }

        return entities;
    }

    this.filterMapPoints = function (mapPoints, ignoreGpsLoss, enableFiltering) {
        // Filter the mapPoints if needed.
        var lastValidLocation = null;

        var orphanedSplitPoint = null;

        var filteredMapPoints = new Array();

        for (var i = 0; i < mapPoints.length; i++) {
            var point = mapPoints[i];
            var location = this.createLocation(point);
            var shouldAddMapPoint = false;
            var isValidPoint = false;

            // If there is an orphaned split marker, and the next map point is not a WayPoint, the orphaned split marker should be discarded
            if (point.MapPointType !== this.constants.wayPoint) {
                orphanedSplitPoint = null;
            }

            if (location) {
                // If latitude and longitude exists, then these should also exists.
                if (this.isValidGpsPoint(point.Location, enableFiltering)) {
                    // The GPS point is valid.
                    shouldAddMapPoint = true;
                    isValidPoint = true;

                    // If there is an orphaned split marker assign it to the next WayPoint with a valid location
                    if (orphanedSplitPoint && point.MapPointType === this.constants.wayPoint) {
                        point.MapPointType = this.constants.splitMarker;
                        point.SplitOrdinal = orphanedSplitPoint.SplitOrdinal;
                        orphanedSplitPoint = null;
                    }

                } else {
                    // The GPS point is not valid. We will need to clear the location.
                    point.Location = null;

                    if (!ignoreGpsLoss) {
                        shouldAddMapPoint = true;
                        isValidPoint = false;
                    }

                    var isSkippingAhead = false;

                    if (lastValidLocation) {
                        //If there is a valid point coming next that's within roughly 100 feet, skip to it to not show short gaps (under a bridge, etc.)
                        for (var j = i + 1; (j < i + this.constants.checkAheadForNearbyPoints) && (j < mapPoints.length) && !isSkippingAhead; j++) {
                            var nextPoint = mapPoints[j];
                            var nextLocation = this.createLocation(nextPoint);

                            //! Microsoft.Maps.Location has lowercase .latitude; our mapPoints[] have Uppercase .Latitude
                            if (nextLocation && this.isValidGpsPoint(nextPoint.Location)) {
                                if (Math.abs(nextLocation.longitude - lastValidLocation.longitude) <= (this.constants.approx100FeetInDegrees) &&
                                    Math.abs(nextLocation.latitude - lastValidLocation.latitude) <= (this.constants.approx100FeetInDegrees)) {
                                    isSkippingAhead = true;
                                    //Change i so i++ when loop ends gets to this point
                                    i = j - 1;
                                }
                            }
                        }
                    }
                }
            } else {

                if (point.MapPointType === this.constants.splitMarker) {
                    orphanedSplitPoint = point;
                }

                // No GPS coordinates.
                if (!ignoreGpsLoss) {
                    // If GPS loss should NOT be ignored add the point to the Array
                    shouldAddMapPoint = true;
                }
            }

            if (shouldAddMapPoint) {
                filteredMapPoints.push(point);
            }

            if (isValidPoint) {
                lastValidLocation = location;
            }
        }

        return filteredMapPoints;
    }

    this.isValidGpsPoint = function (location, enableFiltering) {
        if (enableFiltering) {
            var speedOverGround = location.SpeedOverGround;
            var ehpe = location.EHPE;

            return (speedOverGround >= this.constants.speedThreshold && ehpe < this.constants.gpsErrorThreshold);
        } else {
            return true;
        }
    }

    this.setupHiDPICanvas = function (legendElementId, w, h, ratio) {
        var canvas = document.getElementById(legendElementId);
        var context = canvas.getContext("2d");

        if (!ratio) {
            var devicePixelRatio = window.devicePixelRatio || 1;
            var browserStorePixel = context.webkitBackingStorePixelRatio ||
                context.mozBackingStorePixelRatio ||
                context.msBackingStorePixelRatio ||
                context.oBackingStorePixelRatio ||
                context.backingStorePixelRatio || 1;

            ratio = devicePixelRatio / browserStorePixel;
        }

        canvas.width = w * ratio;
        canvas.height = h * ratio;
        canvas.style.width = w + "px";
        canvas.style.height = h + "px";
        context.setTransform(ratio, 0, 0, ratio, 0, 0);

        return context;
    }

    this.drawLegend = function (legendElementId) {
        var legendWidth = 110;
        var legendHeight = 30;
        var iconSize = 28;
        var margin = 5;
        var lineMargin = 2;
        var segmentWidth = (legendWidth - (margin * 2)) / 3.0;
        var lineY = legendHeight - (margin * 2);
        var lineHeight = 3;

        var context = this.setupHiDPICanvas(legendElementId, legendWidth, legendHeight);

        // Draw the box
        context.fillStyle = "#000000";
        context.fillRect(0, 0, legendWidth, legendHeight);

        // Draw the lines
        context.fillStyle = this.constants.paceColors[0];
        context.fillRect(margin, lineY, segmentWidth, lineHeight);

        context.fillStyle = this.constants.paceColors[1];
        context.fillRect(margin + segmentWidth, lineY, segmentWidth, lineHeight);

        context.fillStyle = this.constants.paceColors[2];
        context.fillRect(margin + (segmentWidth * 2), lineY, segmentWidth, lineHeight);

        // Draw the icons
        var cheetah = new Image();

        cheetah.onload = function () {
            var scaledHeight = (iconSize / cheetah.width) * cheetah.height;
            var y = lineY - lineMargin - scaledHeight;

            context.drawImage(cheetah, legendWidth - (margin + iconSize), y, iconSize, scaledHeight);
        };

        cheetah.src = this.assetPath + 'cheetah.png';

        var snail = new Image();

        snail.onload = function () {
            var scaledHeight = (iconSize / snail.width) * snail.height;
            var y = lineY - lineMargin - scaledHeight;

            context.drawImage(snail, margin, y, iconSize, scaledHeight);
        };

        snail.src = this.assetPath + 'snail.png';
    }
}
